#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""


class EEWeatherError(Exception):
    """Base class for exceptions in the eeweather package."""

    pass


class UnrecognizedUSAFIDError(EEWeatherError):
    """ Raised when an unrecognized USAF station id is encountered.

    Attributes
    ----------
    value : str
        the value which is not a valid USAF ID
    message : str
        a message describing the error
    """

    def __init__(self, value):
        self.value = value
        self.message = (
            'The value "{}" was not recognized as a valid USAF weather station'
            " identifier.".format(value)
        )


class UnrecognizedZCTAError(EEWeatherError):
    """ Raised when an unrecognized ZCTA is encountered.

    Attributes
    ----------
    value : str
        the value which is not a valid ZCTA
    message : str
        a message describing the error
    """

    def __init__(self, value):
        self.value = value
        self.message = 'The value "{}" was not recognized as a valid ZCTA identifier.'.format(
            value
        )


class ISDDataNotAvailableError(EEWeatherError):
    """ Raised when ISD data is not available for a particular station and year.

    Attributes
    ----------
    usaf_id : str
        the USAF ID for which ISD data does not exist.
    year : int
        the year for which ISD data does not exist.
    message : str
        a message describing the error
    """

    def __init__(self, usaf_id, year):
        self.usaf_id = usaf_id
        self.year = year
        self.message = 'ISD data does not exist for station "{}" in year {}.'.format(
            usaf_id, year
        )


class GSODDataNotAvailableError(EEWeatherError):
    """ Raised when GSOD data is not available for a particular station and year.

    Attributes
    ----------
        usaf_id -- The USAF ID for which GSOD data does not exist.
        year -- The year for which GSOD data does not exist.
    """

    def __init__(self, usaf_id, year):
        self.usaf_id = usaf_id
        self.year = year
        self.message = 'GSOD data does not exist for station "{}" in year {}.'.format(
            usaf_id, year
        )


class TMY3DataNotAvailableError(EEWeatherError):
    """ Raised when TMY3 data is not available for a particular station.

    Attributes
    ----------
    usaf_id : str
        the USAF ID for which TMY3 data does not exist.
    message : str
        a message describing the error
    """

    def __init__(self, usaf_id):
        self.usaf_id = usaf_id
        self.message = 'TMY3 data does not exist for station "{}".'.format(usaf_id)


class CZ2010DataNotAvailableError(EEWeatherError):
    """ Raised when CZ2010 data is not available for a particular station.

    Attributes
    ----------
    usaf_id : str
        the USAF ID for which CZ2010 data does not exist.
    message : str
        a message describing the error
    """

    def __init__(self, usaf_id):
        self.usaf_id = usaf_id
        self.message = 'CZ2010 data does not exist for station "{}".'.format(usaf_id)


class NonUTCTimezoneInfoError(EEWeatherError):
    """ Raised when input start and end date aren't explicitly defined 
    to have a UTC timezone.

    Attributes
    ----------
    usaf_id : str
        the USAF ID for which CZ2010 data does not exist.
    message : str
        a message describing the error
    """

    def __init__(self, this_date):
        self.message = (
            '"{}" does not have a UTC timezone. If using the datetime package, it should be'
            " in the format datetime(1,1,1,tzinfo=pytz.UTC).".format(this_date)
        )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_rank_stations_no_filter df.shape"] = (3874, 15)

snapshots["test_rank_stations_match_climate_zones_null match_iecc_climate_zone"] = (
    497,
    15,
)

snapshots["test_rank_stations_match_climate_zones_null match_iecc_moisture_regime"] = (
    933,
    15,
)

snapshots["test_rank_stations_match_climate_zones_null match_ba_climate_zone"] = (
    502,
    15,
)

snapshots["test_rank_stations_match_climate_zones_null match_ca_climate_zone"] = (
    3641,
    15,
)

snapshots["test_rank_stations_match_state site_state=CA, match_state=False"] = (
    3874,
    15,
)

snapshots["test_rank_stations_match_state site_state=CA, match_state=True"] = (306, 15)

snapshots["test_rank_stations_match_state site_state=None, match_state=True"] = (71, 15)

snapshots["test_rank_stations_is_tmy3 is_tmy3=True"] = (1019, 15)

snapshots["test_rank_stations_is_tmy3 is_tmy3=False"] = (2855, 15)

snapshots["test_rank_stations_is_cz2010 is_cz2010=True"] = (86, 15)

snapshots["test_rank_stations_is_cz2010 is_cz2010=False"] = (3788, 15)

snapshots["test_rank_stations_minimum_quality minimum_quality=low"] = (3874, 15)

snapshots["test_rank_stations_minimum_quality minimum_quality=medium"] = (1752, 15)

snapshots["test_rank_stations_minimum_quality minimum_quality=high"] = (1595, 15)

snapshots["test_rank_stations_minimum_tmy3_class minimum_tmy3_class=III"] = (1019, 15)

snapshots["test_rank_stations_minimum_tmy3_class minimum_tmy3_class=II"] = (857, 15)

snapshots["test_rank_stations_minimum_tmy3_class minimum_tmy3_class=I"] = (222, 15)

snapshots["test_rank_stations_max_distance_meters max_distance_meters=200000"] = (
    52,
    15,
)

snapshots["test_rank_stations_max_distance_meters max_distance_meters=50000"] = (5, 15)

snapshots[
    "test_rank_stations_max_difference_elevation_meters max_difference_elevation_meters=200"
] = (3874, 15)

snapshots[
    "test_rank_stations_max_difference_elevation_meters site_elevation=0, max_difference_elevation_meters=200"
] = (2185, 15)

snapshots[
    "test_rank_stations_max_difference_elevation_meters site_elevation=0, max_difference_elevation_meters=50"
] = (1366, 15)

snapshots[
    "test_rank_stations_max_difference_elevation_meters site_elevation=1000, max_difference_elevation_meters=50"
] = (46, 15)

snapshots["test_rank_stations_match_climate_zones_not_null match_iecc_climate_zone"] = (
    743,
    15,
)

snapshots[
    "test_rank_stations_match_climate_zones_not_null match_iecc_moisture_regime"
] = (710, 15)

snapshots["test_rank_stations_match_climate_zones_not_null match_ba_climate_zone"] = (
    280,
    15,
)

snapshots["test_rank_stations_match_climate_zones_not_null match_ca_climate_zone"] = (
    10,
    15,
)

snapshots["test_select_station_with_second_level_dates station_id"] = "723895"

snapshots["test_select_station_with_empty_tempC station_id"] = "723896"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json
from click.testing import CliRunner


from eeweather.cli import (
    cli,
    inspect_isd_station,
    inspect_isd_file_years,
    inspect_isd_filenames,
    inspect_gsod_filenames,
)


def test_eeweather_cli():
    runner = CliRunner()

    result = runner.invoke(cli)
    assert result.exit_code == 0
    assert len(result.output) > 100


def test_inspect_isd_station():
    runner = CliRunner()

    result = runner.invoke(inspect_isd_station, ["722880"])
    assert result.exit_code == 0
    assert json.loads(result.output) == {
        "ba_climate_zone": "Hot-Dry",
        "ca_climate_zone": "CA_09",
        "elevation": "+0236.2",
        "icao_code": "KBUR",
        "iecc_climate_zone": "3",
        "iecc_moisture_regime": "B",
        "latitude": "+34.201",
        "longitude": "-118.358",
        "name": "BURBANK-GLENDALE-PASA ARPT",
        "quality": "high",
        "recent_wban_id": "23152",
        "state": "CA",
        "usaf_id": "722880",
        "wban_ids": "23152,99999",
    }


def test_inspect_isd_station_unrecognized():
    runner = CliRunner()
    result = runner.invoke(inspect_isd_station, ["INVALID"])
    assert result.exit_code == 1


def test_inspect_isd_file_years():
    runner = CliRunner()

    result = runner.invoke(inspect_isd_file_years, ["722880"])
    assert result.exit_code == 0
    assert json.loads(result.output) == [
        {"usaf_id": "722880", "wban_id": "23152", "year": "2006"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2007"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2008"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2009"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2010"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2011"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2012"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2013"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2014"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2015"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2016"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2017"},
        {"usaf_id": "722880", "wban_id": "23152", "year": "2018"},
    ]


def test_inspect_isd_file_years_unrecognized():
    runner = CliRunner()
    result = runner.invoke(inspect_isd_file_years, ["INVALID"])
    assert result.exit_code == 1


def test_inspect_isd_filenames():
    runner = CliRunner()

    result = runner.invoke(inspect_isd_filenames, ["722880", "2017"])
    assert result.exit_code == 0
    assert result.output == (
        "ftp://ftp.ncdc.noaa.gov/pub/data/noaa/2017/722880-23152-2017.gz\n"
    )


def test_inspect_isd_filenames_unrecognized():
    runner = CliRunner()
    result = runner.invoke(inspect_isd_filenames, ["INVALID", "2017"])
    assert result.exit_code == 1


def test_inspect_gsod_filenames():
    runner = CliRunner()

    result = runner.invoke(inspect_gsod_filenames, ["722880", "2017"])
    assert result.exit_code == 0
    assert result.output == (
        "ftp://ftp.ncdc.noaa.gov/pub/data/gsod/2017/722880-23152-2017.op.gz\n"
    )


def test_inspect_gsod_filenames_unrecognized():
    runner = CliRunner()
    result = runner.invoke(inspect_gsod_filenames, ["INVALID", "2017"])
    assert result.exit_code == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from collections import defaultdict
from datetime import datetime, timedelta
import json
import logging
import os
import shutil
import subprocess
import tempfile

import pandas as pd
import numpy as np

from .connections import noaa_ftp_connection_proxy, metadata_db_connection_proxy


logger = logging.getLogger(__name__)

__all__ = ("build_metadata_db", "inspect_metadata_db")

CZ2010_LIST = [
    "725958",
    "725945",
    "723840",
    "724837",
    "724800",
    "725845",
    "747188",
    "722880",
    "723926",
    "722926",
    "722927",
    "746120",
    "722899",
    "724936",
    "725946",
    "723815",
    "723810",
    "722810",
    "725940",
    "723890",
    "722976",
    "724935",
    "747185",
    "722909",
    "723826",
    "722956",
    "725847",
    "723816",
    "747020",
    "724927",
    "722895",
    "722970",
    "722975",
    "722874",
    "722950",
    "724815",
    "724926",
    "722953",
    "725955",
    "724915",
    "725957",
    "724955",
    "723805",
    "724930",
    "723927",
    "722868",
    "747187",
    "723820",
    "724937",
    "723965",
    "723910",
    "723895",
    "725910",
    "725920",
    "722860",
    "722869",
    "724830",
    "724839",
    "724917",
    "724938",
    "722925",
    "722907",
    "722900",
    "722903",
    "722906",
    "724940",
    "724945",
    "724946",
    "722897",
    "722910",
    "723830",
    "722977",
    "723925",
    "723940",
    "722885",
    "724957",
    "724920",
    "722955",
    "745160",
    "725846",
    "690150",
    "725905",
    "722886",
    "723930",
    "723896",
    "724838",
]


class PrettyFloat(float):
    def __repr__(self):
        return "%.7g" % self


def pretty_floats(obj):
    if isinstance(obj, float):
        return PrettyFloat(round(obj, 4))
    elif isinstance(obj, dict):
        return dict((k, pretty_floats(v)) for k, v in obj.items())
    elif isinstance(obj, (list, tuple)):
        return list(map(pretty_floats, obj))
    return obj


def to_geojson(polygon):
    import simplejson
    from shapely.geometry import mapping

    return simplejson.dumps(pretty_floats(mapping(polygon)), separators=(",", ":"))


def _download_primary_sources():
    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    scripts_path = os.path.join(root_dir, "scripts", "download_primary_sources.sh")

    download_path = tempfile.mkdtemp()
    subprocess.call([scripts_path, download_path])
    return download_path


def _load_isd_station_metadata(download_path):
    """ Collect metadata for US isd stations.
    """
    from shapely.geometry import Point

    # load ISD history which contains metadata
    isd_history = pd.read_csv(
        os.path.join(download_path, "isd-history.csv"),
        dtype=str,
        parse_dates=["BEGIN", "END"],
    )

    hasGEO = (
        isd_history.LAT.notnull() & isd_history.LON.notnull() & (isd_history.LAT != 0)
    )
    isUS = (
        ((isd_history.CTRY == "US") & (isd_history.STATE.notnull()))
        # AQ = American Samoa, GQ = Guam, RQ = Peurto Rico, VQ = Virgin Islands
        | (isd_history.CTRY.str[1] == "Q")
    )
    hasUSAF = isd_history.USAF != "999999"

    metadata = {}
    for usaf_station, group in isd_history[hasGEO & isUS & hasUSAF].groupby("USAF"):
        # find most recent
        recent = group.loc[group.END.idxmax()]
        wban_stations = list(group.WBAN)
        metadata[usaf_station] = {
            "usaf_id": usaf_station,
            "wban_ids": wban_stations,
            "recent_wban_id": recent.WBAN,
            "name": recent["STATION NAME"],
            "icao_code": recent.ICAO,
            "latitude": recent.LAT if recent.LAT not in ("+00.000",) else None,
            "longitude": recent.LON if recent.LON not in ("+000.000",) else None,
            "point": Point(float(recent.LON), float(recent.LAT)),
            "elevation": recent["ELEV(M)"]
            if not str(float(recent["ELEV(M)"])).startswith("-999")
            else None,
            "state": recent.STATE,
        }

    return metadata


def _load_isd_file_metadata(download_path, isd_station_metadata):
    """ Collect data counts for isd files.
    """

    isd_inventory = pd.read_csv(
        os.path.join(download_path, "isd-inventory.csv"), dtype=str
    )

    # filter to stations with metadata
    station_keep = [usaf in isd_station_metadata for usaf in isd_inventory.USAF]
    isd_inventory = isd_inventory[station_keep]
    # filter by year
    year_keep = isd_inventory.YEAR > "2005"
    isd_inventory = isd_inventory[year_keep]

    metadata = {}
    for (usaf_station, year), group in isd_inventory.groupby(["USAF", "YEAR"]):
        if usaf_station not in metadata:
            metadata[usaf_station] = {"usaf_id": usaf_station, "years": {}}
        metadata[usaf_station]["years"][year] = [
            {
                "wban_id": row.WBAN,
                "counts": [
                    row.JAN,
                    row.FEB,
                    row.MAR,
                    row.APR,
                    row.MAY,
                    row.JUN,
                    row.JUL,
                    row.AUG,
                    row.SEP,
                    row.OCT,
                    row.NOV,
                    row.DEC,
                ],
            }
            for i, row in group.iterrows()
        ]
    return metadata


def _compute_isd_station_quality(
    isd_station_metadata,
    isd_file_metadata,
    end_year=None,
    years_back=None,
    quality_func=None,
):

    if end_year is None:
        end_year = datetime.now().year - 1  # last full year

    if years_back is None:
        years_back = 5

    if quality_func is None:

        def quality_func(values):
            minimum = values.min()
            if minimum > 24 * 25:
                return "high"
            elif minimum > 24 * 15:
                return "medium"
            else:
                return "low"

    # e.g., if end_year == 2017, year_range = ["2013", "2014", ..., "2017"]
    year_range = set([str(y) for y in range(end_year - (years_back - 1), end_year + 1)])

    def _compute_station_quality(usaf_id):
        years_data = isd_file_metadata.get(usaf_id, {}).get("years", {})
        if not all([year in years_data for year in year_range]):
            return quality_func(np.repeat(0, 60))
        counts = defaultdict(lambda: 0)
        for y, year in enumerate(year_range):
            for station in years_data[year]:
                for m, month_counts in enumerate(station["counts"]):
                    counts[y * 12 + m] += int(month_counts)
        return quality_func(np.array(list(counts.values())))

    # figure out counts for years of interest
    for usaf_id, metadata in isd_station_metadata.items():
        metadata["quality"] = _compute_station_quality(usaf_id)


def _load_zcta_metadata(download_path):
    from shapely.geometry import shape

    # load zcta geojson
    geojson_path = os.path.join(download_path, "cb_2016_us_zcta510_500k.json")
    with open(geojson_path, "r") as f:
        geojson = json.load(f)

    # load ZIP code prefixes by state
    zipcode_prefixes_path = os.path.join(download_path, "zipcode_prefixes.json")
    with open(zipcode_prefixes_path, "r") as f:
        zipcode_prefixes = json.load(f)
        prefix_to_zipcode = {
            zipcode_prefix: state
            for state, zipcode_prefix_list in zipcode_prefixes.items()
            for zipcode_prefix in zipcode_prefix_list
        }

    def _get_state(zcta):
        prefix = zcta[:3]
        return prefix_to_zipcode.get(prefix)

    metadata = {}
    for feature in geojson["features"]:
        zcta = feature["properties"]["GEOID10"]
        geometry = feature["geometry"]
        polygon = shape(geometry)
        centroid = polygon.centroid
        state = _get_state(zcta)
        metadata[zcta] = {
            "zcta": zcta,
            "polygon": polygon,
            "geometry": to_geojson(polygon),
            "centroid": centroid,
            "latitude": centroid.coords[0][1],
            "longitude": centroid.coords[0][0],
            "state": state,
        }
    return metadata


def _load_county_metadata(download_path):
    from shapely.geometry import shape

    # load county geojson
    geojson_path = os.path.join(download_path, "cb_2016_us_county_500k.json")
    with open(geojson_path, "r") as f:
        geojson = json.load(f)

    metadata = {}
    for feature in geojson["features"]:
        county = feature["properties"]["GEOID"]
        geometry = feature["geometry"]
        polygon = shape(geometry)
        centroid = polygon.centroid
        metadata[county] = {
            "county": county,
            "polygon": polygon,
            "geometry": to_geojson(polygon),
            "centroid": centroid,
            "latitude": centroid.coords[0][1],
            "longitude": centroid.coords[0][0],
        }

    # load county climate zones
    county_climate_zones = pd.read_csv(
        os.path.join(download_path, "climate_zones.csv"),
        dtype=str,
        usecols=[
            "State FIPS",
            "County FIPS",
            "IECC Climate Zone",
            "IECC Moisture Regime",
            "BA Climate Zone",
            "County Name",
        ],
    )

    for i, row in county_climate_zones.iterrows():

        county = row["State FIPS"] + row["County FIPS"]
        if county not in metadata:
            logger.warn(
                "Could not find geometry for county {}, skipping.".format(county)
            )
            continue

        metadata[county].update(
            {
                "name": row["County Name"],
                "iecc_climate_zone": row["IECC Climate Zone"],
                "iecc_moisture_regime": (
                    row["IECC Moisture Regime"]
                    if not pd.isnull(row["IECC Moisture Regime"])
                    else None
                ),
                "ba_climate_zone": row["BA Climate Zone"],
            }
        )

    return metadata


def _load_CA_climate_zone_metadata(download_path):
    from shapely.geometry import shape, mapping

    ca_climate_zone_names = {
        "01": "Arcata",
        "02": "Santa Rosa",
        "03": "Oakland",
        "04": "San Jose-Reid",
        "05": "Santa Maria",
        "06": "Torrance",
        "07": "San Diego-Lindbergh",
        "08": "Fullerton",
        "09": "Burbank-Glendale",
        "10": "Riverside",
        "11": "Red Bluff",
        "12": "Sacramento",
        "13": "Fresno",
        "14": "Palmdale",
        "15": "Palm Spring-Intl",
        "16": "Blue Canyon",
    }
    geojson_path = os.path.join(
        download_path, "CA_Building_Standards_Climate_Zones.json"
    )
    with open(geojson_path, "r") as f:
        geojson = json.load(f)

    metadata = {}
    for feature in geojson["features"]:
        zone = "{:02d}".format(int(feature["properties"]["Zone"]))
        geometry = feature["geometry"]
        polygon = shape(geometry)
        metadata[zone] = {
            "ca_climate_zone": "CA_{}".format(zone),
            "name": ca_climate_zone_names[zone],
            "polygon": polygon,
            "geometry": to_geojson(polygon),
        }
    return metadata


def _load_tmy3_station_metadata(download_path):
    from bs4 import BeautifulSoup

    path = os.path.join(download_path, "tmy3-stations.html")
    with open(path, "r") as f:
        soup = BeautifulSoup(f.read(), "html.parser")
    tmy3_station_elements = soup.select("td .hide")

    metadata = {}
    for station_el in tmy3_station_elements:
        station_name_el = station_el.findNext("td").findNext("td")
        station_class_el = station_name_el.findNext("td")
        usaf_id = station_el.text.strip()
        name = (
            "".join(station_name_el.text.split(",")[:-1])
            .replace("\n", "")
            .replace("\t", "")
            .strip()
        )
        metadata[usaf_id] = {
            "usaf_id": usaf_id,
            "name": name,
            "state": station_name_el.text.split(",")[-1].strip(),
            "class": station_class_el.text.split()[1].strip(),
        }
    return metadata


def _load_cz2010_station_metadata():
    return {usaf_id: {"usaf_id": usaf_id} for usaf_id in CZ2010_LIST}


def _create_merged_climate_zones_metadata(county_metadata):
    from shapely.ops import cascaded_union

    iecc_climate_zone_polygons = defaultdict(list)
    iecc_moisture_regime_polygons = defaultdict(list)
    ba_climate_zone_polygons = defaultdict(list)

    for county in county_metadata.values():
        polygon = county["polygon"]
        iecc_climate_zone = county.get("iecc_climate_zone")
        iecc_moisture_regime = county.get("iecc_moisture_regime")
        ba_climate_zone = county.get("ba_climate_zone")
        if iecc_climate_zone is not None:
            iecc_climate_zone_polygons[iecc_climate_zone].append(polygon)
        if iecc_moisture_regime is not None:
            iecc_moisture_regime_polygons[iecc_moisture_regime].append(polygon)
        if ba_climate_zone is not None:
            ba_climate_zone_polygons[ba_climate_zone].append(polygon)

    iecc_climate_zone_metadata = {}
    for iecc_climate_zone, polygons in iecc_climate_zone_polygons.items():
        polygon = cascaded_union(polygons)
        polygon = polygon.simplify(0.01)
        iecc_climate_zone_metadata[iecc_climate_zone] = {
            "iecc_climate_zone": iecc_climate_zone,
            "polygon": polygon,
            "geometry": to_geojson(polygon),
        }

    iecc_moisture_regime_metadata = {}
    for iecc_moisture_regime, polygons in iecc_moisture_regime_polygons.items():
        polygon = cascaded_union(polygons)
        polygon = polygon.simplify(0.01)
        iecc_moisture_regime_metadata[iecc_moisture_regime] = {
            "iecc_moisture_regime": iecc_moisture_regime,
            "polygon": polygon,
            "geometry": to_geojson(polygon),
        }

    ba_climate_zone_metadata = {}
    for ba_climate_zone, polygons in ba_climate_zone_polygons.items():
        polygon = cascaded_union(polygons)
        polygon = polygon.simplify(0.01)
        ba_climate_zone_metadata[ba_climate_zone] = {
            "ba_climate_zone": ba_climate_zone,
            "polygon": polygon,
            "geometry": to_geojson(polygon),
        }

    return (
        iecc_climate_zone_metadata,
        iecc_moisture_regime_metadata,
        ba_climate_zone_metadata,
    )


def _compute_containment(
    point_metadata, point_id_field, polygon_metadata, polygon_metadata_field
):
    from shapely.vectorized import contains

    points, lats, lons = zip(
        *[
            (point, point["latitude"], point["longitude"])
            for point in point_metadata.values()
        ]
    )

    for i, polygon in enumerate(polygon_metadata.values()):
        containment = contains(polygon["polygon"], lons, lats)
        for point, c in zip(points, containment):
            if c:
                point[polygon_metadata_field] = polygon[polygon_metadata_field]
    # fill in with None
    for point in point_metadata.values():
        point[polygon_metadata_field] = point.get(polygon_metadata_field, None)


def _map_zcta_to_climate_zones(
    zcta_metadata,
    iecc_climate_zone_metadata,
    iecc_moisture_regime_metadata,
    ba_climate_zone_metadata,
    ca_climate_zone_metadata,
):

    _compute_containment(
        zcta_metadata, "zcta", iecc_climate_zone_metadata, "iecc_climate_zone"
    )

    _compute_containment(
        zcta_metadata, "zcta", iecc_moisture_regime_metadata, "iecc_moisture_regime"
    )

    _compute_containment(
        zcta_metadata, "zcta", ba_climate_zone_metadata, "ba_climate_zone"
    )

    _compute_containment(
        zcta_metadata, "zcta", ca_climate_zone_metadata, "ca_climate_zone"
    )


def _map_isd_station_to_climate_zones(
    isd_station_metadata,
    iecc_climate_zone_metadata,
    iecc_moisture_regime_metadata,
    ba_climate_zone_metadata,
    ca_climate_zone_metadata,
):

    _compute_containment(
        isd_station_metadata, "usaf_id", iecc_climate_zone_metadata, "iecc_climate_zone"
    )

    _compute_containment(
        isd_station_metadata,
        "usaf_id",
        iecc_moisture_regime_metadata,
        "iecc_moisture_regime",
    )

    _compute_containment(
        isd_station_metadata, "usaf_id", ba_climate_zone_metadata, "ba_climate_zone"
    )

    _compute_containment(
        isd_station_metadata, "usaf_id", ca_climate_zone_metadata, "ca_climate_zone"
    )


def _find_zcta_closest_isd_stations(zcta_metadata, isd_station_metadata, limit=None):
    if limit is None:
        limit = 10
    import pyproj

    geod = pyproj.Geod(ellps="WGS84")

    isd_usaf_ids, isd_lats, isd_lngs = zip(
        *[
            (
                isd_station["usaf_id"],
                float(isd_station["latitude"]),
                float(isd_station["longitude"]),
            )
            for isd_station in isd_station_metadata.values()
        ]
    )

    isd_lats = np.array(isd_lats)
    isd_lngs = np.array(isd_lngs)

    for zcta in zcta_metadata.values():
        zcta_lats = np.tile(zcta["latitude"], isd_lats.shape)
        zcta_lngs = np.tile(zcta["longitude"], isd_lngs.shape)

        dists = geod.inv(zcta_lngs, zcta_lats, isd_lngs, isd_lats)[2]
        sorted_dists = np.argsort(dists)[:limit]

        closest_isd_stations = []
        for i, idx in enumerate(sorted_dists):
            usaf_id = isd_usaf_ids[idx]
            isd_station = isd_station_metadata[usaf_id]
            closest_isd_stations.append(
                {
                    "usaf_id": usaf_id,
                    "distance_meters": int(round(dists[idx])),
                    "rank": i + 1,
                    "iecc_climate_zone_match": (
                        zcta.get("iecc_climate_zone")
                        == isd_station.get("iecc_climate_zone")
                    ),
                    "iecc_moisture_regime_match": (
                        zcta.get("iecc_moisture_regime")
                        == isd_station.get("iecc_moisture_regime")
                    ),
                    "ba_climate_zone_match": (
                        zcta.get("ba_climate_zone")
                        == isd_station.get("ba_climate_zone")
                    ),
                    "ca_climate_zone_match": (
                        zcta.get("ca_climate_zone")
                        == isd_station.get("ca_climate_zone")
                    ),
                }
            )
        zcta["closest_isd_stations"] = closest_isd_stations


def _create_table_structures(conn):
    cur = conn.cursor()
    cur.execute(
        """
      create table isd_station_metadata (
        usaf_id text not null
        , wban_ids text not null
        , recent_wban_id text not null
        , name text not null
        , icao_code text
        , latitude text
        , longitude text
        , elevation text
        , state text
        , quality text default 'low'
        , iecc_climate_zone text
        , iecc_moisture_regime text
        , ba_climate_zone text
        , ca_climate_zone text
      )
    """
    )

    cur.execute(
        """
      create table isd_file_metadata (
        usaf_id text not null
        , year text not null
        , wban_id text not null
      )
    """
    )

    cur.execute(
        """
      create table zcta_metadata (
        zcta_id text not null
        , geometry text
        , latitude text not null
        , longitude text not null
        , state text
        , iecc_climate_zone text
        , iecc_moisture_regime text
        , ba_climate_zone text
        , ca_climate_zone text
      )
    """
    )

    cur.execute(
        """
      create table iecc_climate_zone_metadata (
        iecc_climate_zone text not null
        , geometry text
      )
    """
    )

    cur.execute(
        """
      create table iecc_moisture_regime_metadata (
        iecc_moisture_regime text not null
        , geometry text
      )
    """
    )

    cur.execute(
        """
      create table ba_climate_zone_metadata (
        ba_climate_zone text not null
        , geometry text
      )
    """
    )

    cur.execute(
        """
      create table ca_climate_zone_metadata (
        ca_climate_zone text not null
        , name text not null
        , geometry text
      )
    """
    )

    cur.execute(
        """
      create table tmy3_station_metadata (
        usaf_id text not null
        , name text not null
        , state text not null
        , class text not null
      )
    """
    )

    cur.execute(
        """
      create table cz2010_station_metadata (
        usaf_id text not null
      )
    """
    )


def _write_isd_station_metadata_table(conn, isd_station_metadata):
    cur = conn.cursor()

    rows = [
        (
            metadata["usaf_id"],
            ",".join(metadata["wban_ids"]),
            metadata["recent_wban_id"],
            metadata["name"],
            metadata["icao_code"],
            metadata["latitude"],
            metadata["longitude"],
            metadata["elevation"],
            metadata["state"],
            metadata["quality"],
            metadata["iecc_climate_zone"],
            metadata["iecc_moisture_regime"],
            metadata["ba_climate_zone"],
            metadata["ca_climate_zone"],
        )
        for station, metadata in sorted(isd_station_metadata.items())
    ]
    cur.executemany(
        """
      insert into isd_station_metadata(
        usaf_id
        , wban_ids
        , recent_wban_id
        , name
        , icao_code
        , latitude
        , longitude
        , elevation
        , state
        , quality
        , iecc_climate_zone
        , iecc_moisture_regime
        , ba_climate_zone
        , ca_climate_zone
      ) values (?,?,?,?,?,?,?,?,?,?,?,?,?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index isd_station_metadata_usaf_id on isd_station_metadata(usaf_id)
    """
    )
    cur.execute(
        """
      create index isd_station_metadata_state on isd_station_metadata(state)
    """
    )
    cur.execute(
        """
      create index isd_station_metadata_iecc_climate_zone on
        isd_station_metadata(iecc_climate_zone)
    """
    )
    cur.execute(
        """
      create index isd_station_metadata_iecc_moisture_regime on
        isd_station_metadata(iecc_moisture_regime)
    """
    )
    cur.execute(
        """
      create index isd_station_metadata_ba_climate_zone on
        isd_station_metadata(ba_climate_zone)
    """
    )
    cur.execute(
        """
      create index isd_station_metadata_ca_climate_zone on
        isd_station_metadata(ca_climate_zone)
    """
    )
    cur.close()
    conn.commit()


def _write_isd_file_metadata_table(conn, isd_file_metadata):
    cur = conn.cursor()

    rows = [
        (metadata["usaf_id"], year, station_data["wban_id"])
        for isd_station, metadata in sorted(isd_file_metadata.items())
        for year, year_data in sorted(metadata["years"].items())
        for station_data in year_data
    ]

    cur.executemany(
        """
      insert into isd_file_metadata(
        usaf_id
        , year
        , wban_id
      ) values (?,?,?)
    """,
        rows,
    )

    cur.execute(
        """
      create index isd_file_metadata_usaf_id on
        isd_file_metadata(usaf_id)
    """
    )
    cur.execute(
        """
      create index isd_file_metadata_year on
        isd_file_metadata(year)
    """
    )
    cur.execute(
        """
      create index isd_file_metadata_usaf_id_year on
        isd_file_metadata(usaf_id, year)
    """
    )
    cur.execute(
        """
      create index isd_file_metadata_wban_id on
        isd_file_metadata(wban_id)
    """
    )

    cur.close()
    conn.commit()


def _write_zcta_metadata_table(conn, zcta_metadata, geometry=False):
    cur = conn.cursor()

    rows = [
        (
            metadata["zcta"],
            metadata["geometry"] if geometry else None,
            metadata["latitude"],
            metadata["longitude"],
            metadata["state"],
            metadata["iecc_climate_zone"],
            metadata["iecc_moisture_regime"],
            metadata["ba_climate_zone"],
            metadata["ca_climate_zone"],
        )
        for zcta, metadata in sorted(zcta_metadata.items())
    ]
    cur.executemany(
        """
      insert into zcta_metadata(
        zcta_id
        , geometry
        , latitude
        , longitude
        , state
        , iecc_climate_zone
        , iecc_moisture_regime
        , ba_climate_zone
        , ca_climate_zone
      ) values (?,?,?,?,?,?,?,?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index zcta_metadata_zcta_id on zcta_metadata(zcta_id)
    """
    )
    cur.execute(
        """
      create index zcta_metadata_state on zcta_metadata(state)
    """
    )
    cur.execute(
        """
      create index zcta_metadata_iecc_climate_zone on zcta_metadata(iecc_climate_zone)
    """
    )
    cur.execute(
        """
      create index zcta_metadata_iecc_moisture_regime on zcta_metadata(iecc_moisture_regime)
    """
    )
    cur.execute(
        """
      create index zcta_metadata_ba_climate_zone on zcta_metadata(ba_climate_zone)
    """
    )
    cur.execute(
        """
      create index zcta_metadata_ca_climate_zone on zcta_metadata(ca_climate_zone)
    """
    )
    cur.close()
    conn.commit()


def _write_iecc_climate_zone_metadata_table(
    conn, iecc_climate_zone_metadata, geometry=True
):
    cur = conn.cursor()

    rows = [
        (metadata["iecc_climate_zone"], metadata["geometry"] if geometry else None)
        for iecc_climate_zone, metadata in sorted(iecc_climate_zone_metadata.items())
    ]
    cur.executemany(
        """
      insert into iecc_climate_zone_metadata(
        iecc_climate_zone
        , geometry
      ) values (?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index iecc_climate_zone_metadata_iecc_climate_zone on
      iecc_climate_zone_metadata(iecc_climate_zone)
    """
    )
    cur.close()
    conn.commit()


def _write_iecc_moisture_regime_metadata_table(
    conn, iecc_moisture_regime_metadata, geometry=True
):
    cur = conn.cursor()

    rows = [
        (metadata["iecc_moisture_regime"], metadata["geometry"] if geometry else None)
        for iecc_moisture_regime, metadata in sorted(
            iecc_moisture_regime_metadata.items()
        )
    ]
    cur.executemany(
        """
      insert into iecc_moisture_regime_metadata(
        iecc_moisture_regime
        , geometry
      ) values (?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index iecc_moisture_regime_metadata_iecc_moisture_regime on
      iecc_moisture_regime_metadata(iecc_moisture_regime)
    """
    )
    cur.close()
    conn.commit()


def _write_ba_climate_zone_metadata_table(
    conn, ba_climate_zone_metadata, geometry=True
):
    cur = conn.cursor()

    rows = [
        (metadata["ba_climate_zone"], metadata["geometry"] if geometry else None)
        for ba_climate_zone, metadata in sorted(ba_climate_zone_metadata.items())
    ]
    cur.executemany(
        """
      insert into ba_climate_zone_metadata(
        ba_climate_zone
        , geometry
      ) values (?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index ba_climate_zone_metadata_ba_climate_zone on
      ba_climate_zone_metadata(ba_climate_zone)
    """
    )
    cur.close()
    conn.commit()


def _write_ca_climate_zone_metadata_table(
    conn, ca_climate_zone_metadata, geometry=True
):
    cur = conn.cursor()

    rows = [
        (
            metadata["ca_climate_zone"],
            metadata["name"],
            metadata["geometry"] if geometry else None,
        )
        for ca_climate_zone, metadata in sorted(ca_climate_zone_metadata.items())
    ]
    cur.executemany(
        """
      insert into ca_climate_zone_metadata(
        ca_climate_zone
        , name
        , geometry
      ) values (?,?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index ca_climate_zone_metadata_ca_climate_zone on
      ca_climate_zone_metadata(ca_climate_zone)
    """
    )
    cur.close()
    conn.commit()


def _write_tmy3_station_metadata_table(conn, tmy3_station_metadata):
    cur = conn.cursor()
    rows = [
        (metadata["usaf_id"], metadata["name"], metadata["state"], metadata["class"])
        for tmy3_station, metadata in sorted(tmy3_station_metadata.items())
    ]
    cur.executemany(
        """
      insert into tmy3_station_metadata(
        usaf_id
        , name
        , state
        , class
      ) values (?,?,?,?)
    """,
        rows,
    )
    cur.execute(
        """
      create index tmy3_station_metadata_usaf_id on
        tmy3_station_metadata(usaf_id)
    """
    )
    cur.close()
    conn.commit()


def _write_cz2010_station_metadata_table(conn, cz2010_station_metadata):
    cur = conn.cursor()
    rows = [
        (metadata["usaf_id"],)
        for cz2010_station, metadata in sorted(cz2010_station_metadata.items())
    ]
    cur.executemany(
        """
      insert into cz2010_station_metadata(
        usaf_id
      ) values (?)
    """,
        rows,
    )
    cur.execute(
        """
      create index cz2010_station_metadata_usaf_id on
        cz2010_station_metadata(usaf_id)
    """
    )
    cur.close()
    conn.commit()


def build_metadata_db(
    zcta_geometry=False,
    iecc_climate_zone_geometry=True,
    iecc_moisture_regime_geometry=True,
    ba_climate_zone_geometry=True,
    ca_climate_zone_geometry=True,
):
    """ Build database of metadata from primary sources.

    Downloads primary sources, clears existing DB, and rebuilds from scratch.

    Parameters
    ----------
    zcta_geometry : bool, optional
        Whether or not to include ZCTA geometry in database.
    iecc_climate_zone_geometry : bool, optional
        Whether or not to include IECC Climate Zone geometry in database.
    iecc_moisture_regime_geometry : bool, optional
        Whether or not to include IECC Moisture Regime geometry in database.
    ba_climate_zone_geometry : bool, optional
        Whether or not to include Building America Climate Zone geometry in database.
    ca_climate_zone_geometry : bool, optional
        Whether or not to include California Building Climate Zone Area geometry in database.
    """

    try:
        import shapely
    except ImportError:
        raise ImportError("Loading polygons requires shapely.")

    try:
        from bs4 import BeautifulSoup
    except ImportError:
        raise ImportError("Scraping TMY3 station data requires beautifulsoup4.")

    try:
        import pyproj
    except ImportError:
        raise ImportError("Computing distances requires pyproj.")

    try:
        import simplejson
    except ImportError:
        raise ImportError("Writing geojson requires simplejson.")

    download_path = _download_primary_sources()
    conn = metadata_db_connection_proxy.reset_database()

    # Load data into memory
    print("Loading ZCTAs")
    zcta_metadata = _load_zcta_metadata(download_path)

    print("Loading counties")
    county_metadata = _load_county_metadata(download_path)

    print("Merging county climate zones")
    (
        iecc_climate_zone_metadata,
        iecc_moisture_regime_metadata,
        ba_climate_zone_metadata,
    ) = _create_merged_climate_zones_metadata(county_metadata)

    print("Loading CA climate zones")
    ca_climate_zone_metadata = _load_CA_climate_zone_metadata(download_path)

    print("Loading ISD station metadata")
    isd_station_metadata = _load_isd_station_metadata(download_path)

    print("Loading ISD station file metadata")
    isd_file_metadata = _load_isd_file_metadata(download_path, isd_station_metadata)

    print("Loading TMY3 station metadata")
    tmy3_station_metadata = _load_tmy3_station_metadata(download_path)

    print("Loading CZ2010 station metadata")
    cz2010_station_metadata = _load_cz2010_station_metadata()

    # Augment data in memory
    print("Computing ISD station quality")
    # add rough station quality to station metadata
    # (all months in last 5 years have at least 600 points)
    _compute_isd_station_quality(isd_station_metadata, isd_file_metadata)

    print("Mapping ZCTAs to climate zones")
    # add county and ca climate zone mappings
    _map_zcta_to_climate_zones(
        zcta_metadata,
        iecc_climate_zone_metadata,
        iecc_moisture_regime_metadata,
        ba_climate_zone_metadata,
        ca_climate_zone_metadata,
    )

    print("Mapping ISD stations to climate zones")
    # add county and ca climate zone mappings
    _map_isd_station_to_climate_zones(
        isd_station_metadata,
        iecc_climate_zone_metadata,
        iecc_moisture_regime_metadata,
        ba_climate_zone_metadata,
        ca_climate_zone_metadata,
    )

    # Write tables
    print("Creating table structures")
    _create_table_structures(conn)

    print("Writing ZCTA data")
    _write_zcta_metadata_table(conn, zcta_metadata, geometry=zcta_geometry)

    print("Writing IECC climate zone data")
    _write_iecc_climate_zone_metadata_table(
        conn, iecc_climate_zone_metadata, geometry=iecc_climate_zone_geometry
    )

    print("Writing IECC moisture regime data")
    _write_iecc_moisture_regime_metadata_table(
        conn, iecc_moisture_regime_metadata, geometry=iecc_moisture_regime_geometry
    )

    print("Writing BA climate zone data")
    _write_ba_climate_zone_metadata_table(
        conn, ba_climate_zone_metadata, geometry=ba_climate_zone_geometry
    )

    print("Writing CA climate zone data")
    _write_ca_climate_zone_metadata_table(
        conn, ca_climate_zone_metadata, geometry=ca_climate_zone_geometry
    )

    print("Writing ISD station metadata")
    _write_isd_station_metadata_table(conn, isd_station_metadata)

    print("Writing ISD file metadata")
    _write_isd_file_metadata_table(conn, isd_file_metadata)

    print("Writing TMY3 station metadata")
    _write_tmy3_station_metadata_table(conn, tmy3_station_metadata)

    print("Writing CZ2010 station metadata")
    _write_cz2010_station_metadata_table(conn, cz2010_station_metadata)

    print("Cleaning up...")
    shutil.rmtree(download_path)

    print("\u2728  Completed! \u2728")


def inspect_metadata_db():
    subprocess.call(["sqlite3", metadata_db_connection_proxy.db_path])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from datetime import datetime, timedelta
import gzip
import json
import pkg_resources
import pandas as pd
import pytz

# this import allows monkeypatching noaa_ftp_connection_proxy in tests because
# the fully qualified package path name is preserved
import requests

from .exceptions import (
    UnrecognizedUSAFIDError,
    ISDDataNotAvailableError,
    GSODDataNotAvailableError,
    TMY3DataNotAvailableError,
    CZ2010DataNotAvailableError,
    NonUTCTimezoneInfoError,
)
from .validation import valid_usaf_id_or_raise
from .warnings import EEWeatherWarning
import eeweather.connections
from eeweather.connections import metadata_db_connection_proxy
import eeweather.mockable

DATA_EXPIRATION_DAYS = 1

__all__ = (
    "ISDStation",
    "get_isd_filenames",
    "get_gsod_filenames",
    "get_isd_station_metadata",
    "get_isd_file_metadata",
    "get_isd_raw_temp_data",  # Not currently written
    "get_isd_hourly_temp_data",  # Not currently written
    "get_isd_daily_temp_data",  # Not currently written
    "get_gsod_raw_temp_data",  # Not currently written
    "get_gsod_daily_temp_data",  # Not currently written
    "get_isd_hourly_temp_data_cache_key",
    "get_isd_daily_temp_data_cache_key",
    "get_gsod_daily_temp_data_cache_key",
    "get_tmy3_hourly_temp_data_cache_key",
    "get_cz2010_hourly_temp_data_cache_key",
    "cached_isd_hourly_temp_data_is_expired",
    "cached_isd_daily_temp_data_is_expired",
    "cached_gsod_daily_temp_data_is_expired",
    "validate_isd_hourly_temp_data_cache",
    "validate_isd_daily_temp_data_cache",
    "validate_gsod_daily_temp_data_cache",
    "validate_tmy3_hourly_temp_data_cache",
    "validate_cz2010_hourly_temp_data_cache",
    "serialize_isd_hourly_temp_data",
    "serialize_isd_daily_temp_data",
    "serialize_gsod_daily_temp_data",
    "serialize_tmy3_hourly_temp_data",
    "serialize_cz2010_hourly_temp_data",
    "deserialize_isd_hourly_temp_data",
    "deserialize_isd_daily_temp_data",
    "deserialize_gsod_daily_temp_data",
    "deserialize_tmy3_daily_temp_data",
    "deserialize_cz2010_daily_temp_data",
    "read_isd_hourly_temp_data_from_cache",
    "read_isd_daily_temp_data_from_cache",
    "read_gsod_daily_temp_data_from_cache",
    "read_tmy3_hourly_temp_data_from_cache",
    "read_cz2010_hourly_temp_data_from_cache",
    "write_isd_hourly_temp_data_to_cache",
    "write_isd_daily_temp_data_to_cache",
    "write_gsod_daily_temp_data_to_cache",
    "write_tmy3_hourly_temp_data_to_cache",
    "write_cz2010_hourly_temp_data_to_cache",
    "destroy_cached_isd_hourly_temp_data",
    "destroy_cached_isd_daily_temp_data",
    "destroy_cached_gsod_daily_temp_data",
    "destroy_cached_tmy3_hourly_temp_data",
    "destroy_cached_cz2010_hourly_temp_data",
    "load_isd_hourly_temp_data_cached_proxy",
    "load_isd_daily_temp_data_cached_proxy",
    "load_gsod_daily_temp_data_cached_proxy",
    "load_tmy3_hourly_temp_data_cached_proxy",
    "load_cz2010_hourly_temp_data_cached_proxy",
    "load_isd_hourly_temp_data",
    "load_isd_daily_temp_data",
    "load_gsod_daily_temp_data",
    "load_tmy3_hourly_temp_data",
    "load_cz2010_hourly_temp_data",
    "load_cached_isd_hourly_temp_data",
    "load_cached_isd_daily_temp_data",
    "load_cached_gsod_daily_temp_data",
    "load_cached_tmy3_hourly_temp_data",
    "load_cached_cz2010_hourly_temp_data",
)


def get_isd_filenames(usaf_id, target_year=None, filename_format=None, with_host=False):
    valid_usaf_id_or_raise(usaf_id)
    if filename_format is None:
        filename_format = "/pub/data/noaa/{year}/{usaf_id}-{wban_id}-{year}.gz"
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()

    if target_year is None:
        # all years
        cur.execute(
            """
          select
            wban_id
            , year
          from
            isd_file_metadata
          where
            usaf_id = ?
          order by
            year
        """,
            (usaf_id,),
        )
    else:
        # single year
        cur.execute(
            """
          select
            wban_id
            , year
          from
            isd_file_metadata
          where
            usaf_id = ? and year = ?
        """,
            (usaf_id, target_year),
        )

    filenames = []
    for (wban_id, year) in cur.fetchall():
        filenames.append(
            filename_format.format(usaf_id=usaf_id, wban_id=wban_id, year=year)
        )

    if len(filenames) == 0 and target_year is not None:
        # fallback - use most recent wban id
        cur.execute(
            """
          select
            recent_wban_id
          from
            isd_station_metadata
          where
            usaf_id = ?
        """,
            (usaf_id,),
        )
        row = cur.fetchone()
        if row is not None:
            filenames.append(
                filename_format.format(
                    usaf_id=usaf_id, wban_id=row[0], year=target_year
                )
            )

    if with_host:
        filenames = ["ftp://ftp.ncdc.noaa.gov{}".format(f) for f in filenames]

    return filenames


def get_gsod_filenames(usaf_id, year=None, with_host=False):
    filename_format = "/pub/data/gsod/{year}/{usaf_id}-{wban_id}-{year}.op.gz"
    return get_isd_filenames(
        usaf_id, year, filename_format=filename_format, with_host=with_host
    )


def get_isd_station_metadata(usaf_id):
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()
    cur.execute(
        """
      select
        *
      from
        isd_station_metadata
      where
        usaf_id = ?
    """,
        (usaf_id,),
    )
    row = cur.fetchone()
    if row is None:
        raise UnrecognizedUSAFIDError(usaf_id)
    return {col[0]: row[i] for i, col in enumerate(cur.description)}


def get_isd_file_metadata(usaf_id):
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()
    cur.execute(
        """
      select
        *
      from
        isd_file_metadata
      where
        usaf_id = ?
    """,
        (usaf_id,),
    )
    rows = cur.fetchall()
    if rows == []:
        raise UnrecognizedUSAFIDError(usaf_id)
    return [{col[0]: row[i] for i, col in enumerate(cur.description)} for row in rows]


def get_tmy3_station_metadata(usaf_id):
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()
    cur.execute(
        """
      select
        *
      from
        tmy3_station_metadata
      where
        usaf_id = ?
    """,
        (usaf_id,),
    )
    row = cur.fetchone()
    if row is None:
        raise TMY3DataNotAvailableError(usaf_id)
    return {col[0]: row[i] for i, col in enumerate(cur.description)}


def get_cz2010_station_metadata(usaf_id):
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()
    cur.execute(
        """
      select
        *
      from
        cz2010_station_metadata
      where
        usaf_id = ?
    """,
        (usaf_id,),
    )
    row = cur.fetchone()
    if row is None:
        raise CZ2010DataNotAvailableError(usaf_id)
    return {col[0]: row[i] for i, col in enumerate(cur.description)}


def fetch_isd_raw_temp_data(usaf_id, year):
    # possible locations of this data, errors if station is not recognized
    filenames = get_isd_filenames(usaf_id, year)

    data = []
    for filename in filenames:

        # using fully-qualified name facilitates monkeypatching
        gzipped = eeweather.connections.noaa_ftp_connection_proxy.read_file_as_bytes(
            filename
        )

        if gzipped is not None:
            f = gzip.GzipFile(fileobj=gzipped)
            for line in f.readlines():
                if line[87:92].decode("utf-8") == "+9999":
                    tempC = float("nan")
                else:
                    tempC = float(line[87:92]) / 10.
                date_str = line[15:27].decode("utf-8")
                dt = pytz.UTC.localize(datetime.strptime(date_str, "%Y%m%d%H%M"))
                data.append([dt, tempC])
            gzipped.close()

    if data == []:
        raise ISDDataNotAvailableError(usaf_id, year)

    dates, temps = zip(*sorted(data))
    ts = pd.Series(temps, index=dates)
    ts = ts.groupby(ts.index).mean()
    return ts


def fetch_isd_hourly_temp_data(usaf_id, year):
    # TODO(philngo): allow swappable resample method
    # TODO(philngo): record data sufficiency warnings
    ts = fetch_isd_raw_temp_data(usaf_id, year)

    # CalTRACK 2.3.3
    return (
        ts.resample("Min")
        .mean()
        .interpolate(method="linear", limit=60, limit_direction="both")
        .resample("H")
        .mean()
    )


def fetch_isd_daily_temp_data(usaf_id, year):
    # TODO(philngo): allow swappable resample method
    # TODO(philngo): record data sufficiency warnings
    ts = fetch_isd_raw_temp_data(usaf_id, year)
    return (
        ts.resample("Min")
        .mean()
        .interpolate(method="linear", limit=60, limit_direction="both")
        .resample("D")
        .mean()
    )


def fetch_gsod_raw_temp_data(usaf_id, year):

    filenames = get_gsod_filenames(usaf_id, year)

    data = []
    for filename in filenames:

        # using fully-qualified name facilitates monkeypatching
        gzipped = eeweather.connections.noaa_ftp_connection_proxy.read_file_as_bytes(
            filename
        )

        if gzipped is not None:
            f = gzip.GzipFile(fileobj=gzipped)
            lines = f.readlines()
            for line in lines[1:]:
                columns = line.split()
                date_str = columns[2].decode("utf-8")
                tempF = float(columns[3])
                tempC = (5. / 9.) * (tempF - 32.)
                dt = pytz.UTC.localize(datetime.strptime(date_str, "%Y%m%d"))
                data.append([dt, tempC])
            gzipped.close()

    if data == []:
        raise GSODDataNotAvailableError(usaf_id, year)

    dates, temps = zip(*sorted(data))
    ts = pd.Series(temps, index=dates)
    ts = ts.groupby(ts.index).mean()
    return ts


def fetch_gsod_daily_temp_data(usaf_id, year):
    ts = fetch_gsod_raw_temp_data(usaf_id, year)
    return ts.resample("D").mean()


def fetch_tmy3_hourly_temp_data(usaf_id):
    url = (
        "http://rredc.nrel.gov/solar/old_data/nsrdb/"
        "1991-2005/data/tmy3/{}TYA.CSV".format(usaf_id)
    )

    # checks that the station has TMY3 data associated with it.
    tmy3_metadata = get_tmy3_station_metadata(usaf_id)

    return fetch_hourly_normalized_temp_data(usaf_id, url, "TMY3")


def fetch_cz2010_hourly_temp_data(usaf_id):
    url = "https://storage.googleapis.com/oee-cz2010/csv/{}_CZ2010.CSV".format(usaf_id)

    # checks that the station has CZ2010 data associated with it.
    cz2010_metadata = get_cz2010_station_metadata(usaf_id)

    return fetch_hourly_normalized_temp_data(usaf_id, url, "CZ2010")


@eeweather.mockable.mockable()
def request_text(url):  # pragma: no cover
    response = requests.get(url)
    if response.ok:
        return response.text
    else:
        raise RuntimeError("Could not find {}.".format(url))


def fetch_hourly_normalized_temp_data(usaf_id, url, source_name):
    index = pd.date_range("1900-01-01 00:00", "1900-12-31 23:00", freq="H", tz=pytz.UTC)
    ts = pd.Series(None, index=index, dtype=float)

    lines = eeweather.mockable.request_text(url).splitlines()

    utc_offset_str = lines[0].split(",")[3]
    utc_offset = timedelta(seconds=3600 * float(utc_offset_str))

    for line in lines[2:]:
        row = line.split(",")
        month = row[0][0:2]
        day = row[0][3:5]
        hour = int(row[1][0:2]) - 1

        # YYYYMMDDHH
        date_string = "1900{}{}{:02d}".format(month, day, hour)

        dt = datetime.strptime(date_string, "%Y%m%d%H") - utc_offset

        # Only a little redundant to make year 1900 again - matters for
        # first or last few hours of the year depending UTC on offset
        dt = pytz.UTC.localize(dt.replace(year=1900))
        temp_C = float(row[31])

        ts[dt] = temp_C

    return ts


def get_isd_hourly_temp_data_cache_key(usaf_id, year):
    return "isd-hourly-{}-{}".format(usaf_id, year)


def get_isd_daily_temp_data_cache_key(usaf_id, year):
    return "isd-daily-{}-{}".format(usaf_id, year)


def get_gsod_daily_temp_data_cache_key(usaf_id, year):
    return "gsod-daily-{}-{}".format(usaf_id, year)


def get_tmy3_hourly_temp_data_cache_key(usaf_id):
    return "tmy3-hourly-{}".format(usaf_id)


def get_cz2010_hourly_temp_data_cache_key(usaf_id):
    return "cz2010-hourly-{}".format(usaf_id)


def _expired(last_updated, year):
    if last_updated is None:
        return True
    expiration_limit = pytz.UTC.localize(
        datetime.now() - timedelta(days=DATA_EXPIRATION_DAYS)
    )
    updated_during_data_year = year == last_updated.year
    return expiration_limit > last_updated and updated_during_data_year


def cached_isd_hourly_temp_data_is_expired(usaf_id, year):
    key = get_isd_hourly_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    last_updated = store.key_updated(key)
    return _expired(last_updated, year)


def cached_isd_daily_temp_data_is_expired(usaf_id, year):
    key = get_isd_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    last_updated = store.key_updated(key)
    return _expired(last_updated, year)


def cached_gsod_daily_temp_data_is_expired(usaf_id, year):
    key = get_gsod_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    last_updated = store.key_updated(key)
    return _expired(last_updated, year)


def validate_isd_hourly_temp_data_cache(usaf_id, year):
    key = get_isd_hourly_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()

    # fail if no key
    if not store.key_exists(key):
        return False

    # check for expired data, fail if so
    if cached_isd_hourly_temp_data_is_expired(usaf_id, year):
        store.clear(key)
        return False

    return True


def validate_isd_daily_temp_data_cache(usaf_id, year):
    key = get_isd_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()

    # fail if no key
    if not store.key_exists(key):
        return False

    # check for expired data, fail if so
    if cached_isd_daily_temp_data_is_expired(usaf_id, year):
        store.clear(key)
        return False

    return True


def validate_gsod_daily_temp_data_cache(usaf_id, year):
    key = get_gsod_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()

    # fail if no key
    if not store.key_exists(key):
        return False

    # check for expired data, fail if so
    if cached_gsod_daily_temp_data_is_expired(usaf_id, year):
        store.clear(key)
        return False

    return True


def validate_tmy3_hourly_temp_data_cache(usaf_id):
    key = get_tmy3_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()

    # fail if no key
    if not store.key_exists(key):
        return False

    return True


def validate_cz2010_hourly_temp_data_cache(usaf_id):
    key = get_cz2010_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()

    # fail if no key
    if not store.key_exists(key):
        return False

    return True


def _serialize(ts, freq):
    if freq == "H":
        dt_format = "%Y%m%d%H"
    elif freq == "D":
        dt_format = "%Y%m%d"
    else:  # pragma: no cover
        raise ValueError('Unrecognized frequency "{}"'.format(freq))

    return [
        [d.strftime(dt_format), round(temp, 4) if pd.notnull(temp) else None]
        for d, temp in ts.iteritems()
    ]


def serialize_isd_hourly_temp_data(ts):
    return _serialize(ts, "H")


def serialize_isd_daily_temp_data(ts):
    return _serialize(ts, "D")


def serialize_gsod_daily_temp_data(ts):
    return _serialize(ts, "D")


def serialize_tmy3_hourly_temp_data(ts):
    return _serialize(ts, "H")


def serialize_cz2010_hourly_temp_data(ts):
    return _serialize(ts, "H")


def _deserialize(data, freq):
    if freq == "H":
        dt_format = "%Y%m%d%H"
    elif freq == "D":
        dt_format = "%Y%m%d"
    else:  # pragma: no cover
        raise ValueError('Unrecognized frequency "{}"'.format(freq))

    dates, values = zip(*data)
    index = pd.to_datetime(dates, format=dt_format, utc=True)
    return (
        pd.Series(values, index=index, dtype=float).sort_index().resample(freq).mean()
    )


def deserialize_isd_hourly_temp_data(data):
    return _deserialize(data, "H")


def deserialize_isd_daily_temp_data(data):
    return _deserialize(data, "D")


def deserialize_gsod_daily_temp_data(data):
    return _deserialize(data, "D")


def deserialize_tmy3_hourly_temp_data(data):
    return _deserialize(data, "H")


def deserialize_cz2010_hourly_temp_data(data):
    return _deserialize(data, "H")


def read_isd_hourly_temp_data_from_cache(usaf_id, year):
    key = get_isd_hourly_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return deserialize_isd_hourly_temp_data(store.retrieve_json(key))


def read_isd_daily_temp_data_from_cache(usaf_id, year):
    key = get_isd_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return deserialize_isd_daily_temp_data(store.retrieve_json(key))


def read_gsod_daily_temp_data_from_cache(usaf_id, year):
    key = get_gsod_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return deserialize_gsod_daily_temp_data(store.retrieve_json(key))


def read_tmy3_hourly_temp_data_from_cache(usaf_id):
    key = get_tmy3_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return deserialize_tmy3_hourly_temp_data(store.retrieve_json(key))


def read_cz2010_hourly_temp_data_from_cache(usaf_id):
    key = get_cz2010_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return deserialize_cz2010_hourly_temp_data(store.retrieve_json(key))


def write_isd_hourly_temp_data_to_cache(usaf_id, year, ts):
    key = get_isd_hourly_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.save_json(key, serialize_isd_hourly_temp_data(ts))


def write_isd_daily_temp_data_to_cache(usaf_id, year, ts):
    key = get_isd_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.save_json(key, serialize_isd_daily_temp_data(ts))


def write_gsod_daily_temp_data_to_cache(usaf_id, year, ts):
    key = get_gsod_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.save_json(key, serialize_gsod_daily_temp_data(ts))


def write_tmy3_hourly_temp_data_to_cache(usaf_id, ts):
    key = get_tmy3_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.save_json(key, serialize_tmy3_hourly_temp_data(ts))


def write_cz2010_hourly_temp_data_to_cache(usaf_id, ts):
    key = get_cz2010_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.save_json(key, serialize_cz2010_hourly_temp_data(ts))


def destroy_cached_isd_hourly_temp_data(usaf_id, year):
    key = get_isd_hourly_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.clear(key)


def destroy_cached_isd_daily_temp_data(usaf_id, year):
    key = get_isd_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.clear(key)


def destroy_cached_gsod_daily_temp_data(usaf_id, year):
    key = get_gsod_daily_temp_data_cache_key(usaf_id, year)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.clear(key)


def destroy_cached_tmy3_hourly_temp_data(usaf_id):
    key = get_tmy3_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.clear(key)


def destroy_cached_cz2010_hourly_temp_data(usaf_id):
    key = get_cz2010_hourly_temp_data_cache_key(usaf_id)
    store = eeweather.connections.key_value_store_proxy.get_store()
    return store.clear(key)


def load_isd_hourly_temp_data_cached_proxy(
    usaf_id, year, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):
    # take from cache?
    data_ok = validate_isd_hourly_temp_data_cache(usaf_id, year)

    if not fetch_from_web and not data_ok:
        raise ISDDataNotAvailableError(usaf_id, year)
    elif fetch_from_web and (not read_from_cache or not data_ok):
        # need to actually fetch the data
        ts = fetch_isd_hourly_temp_data(usaf_id, year)
        if write_to_cache:
            write_isd_hourly_temp_data_to_cache(usaf_id, year, ts)
    else:
        # read_from_cache=True and data_ok=True
        ts = read_isd_hourly_temp_data_from_cache(usaf_id, year)
    return ts


def load_isd_daily_temp_data_cached_proxy(
    usaf_id, year, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):
    # take from cache?
    data_ok = validate_isd_daily_temp_data_cache(usaf_id, year)

    if not fetch_from_web and not data_ok:
        raise ISDDataNotAvailableError(usaf_id, year)
    elif fetch_from_web and (not read_from_cache or not data_ok):
        # need to actually fetch the data
        ts = fetch_isd_daily_temp_data(usaf_id, year)
        if write_to_cache:
            write_isd_daily_temp_data_to_cache(usaf_id, year, ts)
    else:
        # read_from_cache=True and data_ok=True
        ts = read_isd_daily_temp_data_from_cache(usaf_id, year)
    return ts


def load_gsod_daily_temp_data_cached_proxy(
    usaf_id, year, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):
    # take from cache?
    data_ok = validate_gsod_daily_temp_data_cache(usaf_id, year)

    if not fetch_from_web and not data_ok:
        raise GSODDataNotAvailableError(usaf_id, year)
    elif fetch_from_web and (not read_from_cache or not data_ok):
        # need to actually fetch the data
        ts = fetch_gsod_daily_temp_data(usaf_id, year)
        if write_to_cache:
            write_gsod_daily_temp_data_to_cache(usaf_id, year, ts)
    else:
        # read_from_cache=True and data_ok=True
        ts = read_gsod_daily_temp_data_from_cache(usaf_id, year)
    return ts


def load_tmy3_hourly_temp_data_cached_proxy(
    usaf_id, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):
    # take from cache?
    data_ok = validate_tmy3_hourly_temp_data_cache(usaf_id)

    if not fetch_from_web and not data_ok:
        raise TMY3DataNotAvailableError(usaf_id)
    elif fetch_from_web and (not read_from_cache or not data_ok):
        # need to actually fetch the data
        ts = fetch_tmy3_hourly_temp_data(usaf_id)
        if write_to_cache:
            write_tmy3_hourly_temp_data_to_cache(usaf_id, ts)
    else:
        # read_from_cache=True and data_ok=True
        ts = read_tmy3_hourly_temp_data_from_cache(usaf_id)
    return ts


def load_cz2010_hourly_temp_data_cached_proxy(
    usaf_id, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):
    # take from cache?
    data_ok = validate_cz2010_hourly_temp_data_cache(usaf_id)

    if not fetch_from_web and not data_ok:
        raise CZ2010DataNotAvailableError(usaf_id)
    elif fetch_from_web and (not read_from_cache or not data_ok):
        # need to actually fetch the data
        ts = fetch_cz2010_hourly_temp_data(usaf_id)
        if write_to_cache:
            write_cz2010_hourly_temp_data_to_cache(usaf_id, ts)
    else:
        # read_from_cache=True and data_ok=True
        ts = read_cz2010_hourly_temp_data_from_cache(usaf_id)
    return ts


def load_isd_hourly_temp_data(
    usaf_id,
    start,
    end,
    read_from_cache=True,
    write_to_cache=True,
    error_on_missing_years=False,
    fetch_from_web=True,
):

    warnings = []
    # CalTRACK 2.3.3
    if start.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    if end.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    if not error_on_missing_years:
        data = []
        for year in range(start.year, end.year + 1):
            try:
                data.append(
                    load_isd_hourly_temp_data_cached_proxy(
                        usaf_id,
                        year,
                        read_from_cache=read_from_cache,
                        write_to_cache=write_to_cache,
                        fetch_from_web=fetch_from_web,
                    )
                )
            except ISDDataNotAvailableError:
                warnings.append(
                    EEWeatherWarning(
                        qualified_name="eeweather.isd_data_not_available",
                        description=("ISD Data not available"),
                        data={"year": year},
                    )
                )
                pass
    else:
        data = [
            load_isd_hourly_temp_data_cached_proxy(
                usaf_id,
                year,
                read_from_cache=read_from_cache,
                write_to_cache=write_to_cache,
                fetch_from_web=fetch_from_web,
            )
            for year in range(start.year, end.year + 1)
        ]

    # get raw data from loaded years into hourly form
    ts = pd.concat(data).resample("H").mean()

    # whittle down to desired range
    ts = ts[start:end]

    if len(ts) > 0:
        # because start and end dates need.to fall exactly on hours
        ts_start = datetime(
            start.year, start.month, start.day, start.hour, tzinfo=pytz.UTC
        )
        # add an hour if not already exactly on an hour, which guarantees
        # that ts_start is greater than or equal to start.
        if ts_start < start:
            ts_start += timedelta(seconds=3600)
        ts_end = datetime(end.year, end.month, end.day, end.hour, tzinfo=pytz.UTC)
        # fill in gaps
        ts = ts.reindex(pd.date_range(ts_start, ts_end, freq="H", tz=pytz.UTC))
    return ts, warnings


def load_isd_daily_temp_data(
    usaf_id, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):

    # CalTRACK 2.3.3
    if start.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    if end.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    data = [
        load_isd_daily_temp_data_cached_proxy(
            usaf_id,
            year,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
        )
        for year in range(start.year, end.year + 1)
    ]

    # get raw data
    ts = pd.concat(data).resample("D").mean()

    # whittle down
    ts = ts[start:end]

    if len(ts) > 0:
        # because start and end dates need.to fall exactly on days
        ts_start = datetime(start.year, start.month, start.day, tzinfo=pytz.UTC)
        # add a day if not already exactly on a day, which guarantees
        # that ts_start is greater than or equal to start.
        if ts_start < start:
            ts_start += timedelta(days=1)
        ts_end = datetime(end.year, end.month, end.day, tzinfo=pytz.UTC)
        # fill in gaps
        ts = ts.reindex(pd.date_range(ts_start, ts_end, freq="D", tz=pytz.UTC))
    return ts


def load_gsod_daily_temp_data(
    usaf_id, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):

    # CalTRACK 2.3.3
    if start.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    if end.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    data = [
        load_gsod_daily_temp_data_cached_proxy(
            usaf_id,
            year,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
        )
        for year in range(start.year, end.year + 1)
    ]
    # get raw data
    ts = pd.concat(data).resample("D").mean()

    # whittle down
    ts = ts[start:end]

    if len(ts) > 0:
        # because start and end dates need.to fall exactly on days
        ts_start = datetime(start.year, start.month, start.day, tzinfo=pytz.UTC)
        # add a day if not already exactly on a day, which guarantees
        # that ts_start is greater than or equal to start.
        if ts_start < start:
            ts_start += timedelta(days=1)
        ts_end = datetime(end.year, end.month, end.day, tzinfo=pytz.UTC)
        # fill in gaps
        ts = ts.reindex(pd.date_range(ts_start, ts_end, freq="D", tz=pytz.UTC))
    return ts


def load_tmy3_hourly_temp_data(
    usaf_id, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):

    # CalTRACK 2.3.3
    if start.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    if end.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    single_year_data = load_tmy3_hourly_temp_data_cached_proxy(
        usaf_id,
        read_from_cache=read_from_cache,
        write_to_cache=write_to_cache,
        fetch_from_web=fetch_from_web,
    )

    # dealing with year replacement
    data = []
    for year in range(start.year, end.year + 1):
        single_year_index = single_year_data.index.map(lambda t: t.replace(year=year))

        data.append(pd.Series(single_year_data.values, index=single_year_index))

    # get raw data
    ts = pd.concat(data).resample("H").mean()

    # whittle down
    ts = ts[start:end]

    # fill in gaps
    ts = ts.reindex(pd.date_range(start, end, freq="H", tz=pytz.UTC))
    return ts


def load_cz2010_hourly_temp_data(
    usaf_id, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
):

    # CalTRACK 2.3.3
    if start.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    if end.tzinfo != pytz.UTC:
        raise NonUTCTimezoneInfoError(start)
    single_year_data = load_cz2010_hourly_temp_data_cached_proxy(
        usaf_id,
        read_from_cache=read_from_cache,
        write_to_cache=write_to_cache,
        fetch_from_web=fetch_from_web,
    )

    # dealing with year replacement
    data = []
    for year in range(start.year, end.year + 1):
        single_year_index = single_year_data.index.map(lambda t: t.replace(year=year))

        data.append(pd.Series(single_year_data.values, index=single_year_index))

    # get raw data
    ts = pd.concat(data).resample("H").mean()

    # whittle down
    ts = ts[start:end]

    # fill in gaps
    ts = ts.reindex(pd.date_range(start, end, freq="H", tz=pytz.UTC))
    return ts


def load_cached_isd_hourly_temp_data(usaf_id):
    store = eeweather.connections.key_value_store_proxy.get_store()

    data = [
        read_isd_hourly_temp_data_from_cache(usaf_id, year)
        for year in range(2000, datetime.now().year + 1)
        if store.key_exists(get_isd_hourly_temp_data_cache_key(usaf_id, year))
    ]
    if data == []:
        return None
    return pd.concat(data).resample("H").mean()


def load_cached_isd_daily_temp_data(usaf_id):
    store = eeweather.connections.key_value_store_proxy.get_store()

    data = [
        read_isd_daily_temp_data_from_cache(usaf_id, year)
        for year in range(2000, datetime.now().year + 1)
        if store.key_exists(get_isd_daily_temp_data_cache_key(usaf_id, year))
    ]
    if data == []:
        return None
    return pd.concat(data).resample("D").mean()


def load_cached_gsod_daily_temp_data(usaf_id):
    store = eeweather.connections.key_value_store_proxy.get_store()

    data = [
        read_gsod_daily_temp_data_from_cache(usaf_id, year)
        for year in range(2000, datetime.now().year + 1)
        if store.key_exists(get_gsod_daily_temp_data_cache_key(usaf_id, year))
    ]
    if data == []:
        return None
    return pd.concat(data).resample("D").mean()


def load_cached_tmy3_hourly_temp_data(usaf_id):
    store = eeweather.connections.key_value_store_proxy.get_store()

    if store.key_exists(get_tmy3_hourly_temp_data_cache_key(usaf_id)):
        return read_tmy3_hourly_temp_data_from_cache(usaf_id)
    else:
        return None


def load_cached_cz2010_hourly_temp_data(usaf_id):
    store = eeweather.connections.key_value_store_proxy.get_store()

    if store.key_exists(get_cz2010_hourly_temp_data_cache_key(usaf_id)):
        return read_cz2010_hourly_temp_data_from_cache(usaf_id)
    else:
        return None


class ISDStation(object):
    """ A representation of an Integrated Surface Database weather station.

    Contains data about a particular ISD station, as well as methods to pull
    data for this station.

    Parameters
    ----------
    usaf_id : str
        ISD station USAF ID
    load_metatdata : bool, optional
        Whether or not to auto-load metadata for this station

    Attributes
    ----------
    usaf_id : str
        ISD station USAF ID
    iecc_climate_zone : str
        IECC Climate Zone
    iecc_moisture_regime : str
        IECC Moisture Regime
    ba_climate_zone : str
        Building America Climate Zone
    ca_climate_zone : str
        California Building Climate Zone
    elevation : float
        elevation of station
    latitude : float
        latitude of station
    longitude : float
        longitude of station
    coords : tuple of (float, float)
        lat/long coordinates of station
    name : str
        name of the station
    quality : str
        "high", "medium", "low"
    wban_ids : list of str
        list of WBAN IDs, or "99999" which have been used to identify the station.
    recent_wban_id = None
        WBAN ID most recently used to identify the station.
    climate_zones = {}
        dict of all climate zones.
    """

    def __init__(self, usaf_id, load_metadata=True):
        self.usaf_id = usaf_id

        if load_metadata:
            self._load_metadata()
        else:
            valid_usaf_id_or_raise(usaf_id)
            self.iecc_climate_zone = None
            self.iecc_moisture_regime = None
            self.ba_climate_zone = None
            self.ca_climate_zone = None
            self.elevation = None
            self.latitude = None
            self.longitude = None
            self.coords = None
            self.name = None
            self.quality = None
            self.wban_ids = None
            self.recent_wban_id = None
            self.climate_zones = {}

    def __str__(self):
        return self.usaf_id

    def __repr__(self):
        return "ISDStation('{}')".format(self.usaf_id)

    def _load_metadata(self):
        metadata = get_isd_station_metadata(self.usaf_id)

        def _float_or_none(field):
            value = metadata.get(field)
            return None if value is None else float(value)

        self.iecc_climate_zone = metadata.get("iecc_climate_zone")
        self.iecc_moisture_regime = metadata.get("iecc_moisture_regime")
        self.ba_climate_zone = metadata.get("ba_climate_zone")
        self.ca_climate_zone = metadata.get("ca_climate_zone")
        self.icao_code = metadata.get("icao_code")
        self.elevation = _float_or_none("elevation")  # meters
        self.latitude = _float_or_none("latitude")
        self.longitude = _float_or_none("longitude")
        self.coords = (self.latitude, self.longitude)
        self.name = metadata.get("name")
        self.quality = metadata.get("quality")
        self.wban_ids = metadata.get("wban_ids", "").split(",")
        self.recent_wban_id = metadata.get("recent_wban_id")
        self.climate_zones = {
            "iecc_climate_zone": metadata.get("iecc_climate_zone"),
            "iecc_moisture_regime": metadata.get("iecc_moisture_regime"),
            "ba_climate_zone": metadata.get("ba_climate_zone"),
            "ca_climate_zone": metadata.get("ca_climate_zone"),
        }

    def json(self):
        """ Return a JSON-serializeable object containing station metadata."""
        return {
            "elevation": self.elevation,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "icao_code": self.icao_code,
            "name": self.name,
            "quality": self.quality,
            "wban_ids": self.wban_ids,
            "recent_wban_id": self.recent_wban_id,
            "climate_zones": {
                "iecc_climate_zone": self.iecc_climate_zone,
                "iecc_moisture_regime": self.iecc_moisture_regime,
                "ba_climate_zone": self.ba_climate_zone,
                "ca_climate_zone": self.ca_climate_zone,
            },
        }

    def get_isd_filenames(self, year=None, with_host=False):
        """ Get filenames of raw ISD station data. """
        return get_isd_filenames(self.usaf_id, year, with_host=with_host)

    def get_gsod_filenames(self, year=None, with_host=False):
        """ Get filenames of raw GSOD station data. """
        return get_gsod_filenames(self.usaf_id, year, with_host=with_host)

    def get_isd_file_metadata(self):
        """ Get raw file metadata for the station. """
        return get_isd_file_metadata(self.usaf_id)

    # fetch raw data
    def fetch_isd_raw_temp_data(self, year):
        """ Pull raw ISD data for the given year directly from FTP. """
        return fetch_isd_raw_temp_data(self.usaf_id, year)

    def fetch_gsod_raw_temp_data(self, year):
        """ Pull raw GSOD data for the given year directly from FTP. """
        return fetch_gsod_raw_temp_data(self.usaf_id, year)

    # fetch raw data then frequency-normalize
    def fetch_isd_hourly_temp_data(self, year):
        """ Pull raw ISD temperature data for the given year directly from FTP and resample to hourly time series. """
        return fetch_isd_hourly_temp_data(self.usaf_id, year)

    def fetch_isd_daily_temp_data(self, year):
        """ Pull raw ISD temperature data for the given year directly from FTP and resample to daily time series. """
        return fetch_isd_daily_temp_data(self.usaf_id, year)

    def fetch_gsod_daily_temp_data(self, year):
        """ Pull raw GSOD temperature data for the given year directly from FTP and resample to daily time series. """
        return fetch_gsod_daily_temp_data(self.usaf_id, year)

    def fetch_tmy3_hourly_temp_data(self):
        """ Pull hourly TMY3 temperature hourly time series directly from NREL. """
        return fetch_tmy3_hourly_temp_data(self.usaf_id)

    def fetch_cz2010_hourly_temp_data(self):
        """ Pull hourly CZ2010 temperature hourly time series from URL. """
        return fetch_cz2010_hourly_temp_data(self.usaf_id)

    # get key-value store key
    def get_isd_hourly_temp_data_cache_key(self, year):
        """ Get key used to cache resampled hourly ISD temperature data for the given year. """
        return get_isd_hourly_temp_data_cache_key(self.usaf_id, year)

    def get_isd_daily_temp_data_cache_key(self, year):
        """ Get key used to cache resampled daily ISD temperature data for the given year. """
        return get_isd_daily_temp_data_cache_key(self.usaf_id, year)

    def get_gsod_daily_temp_data_cache_key(self, year):
        """ Get key used to cache resampled daily GSOD temperature data for the given year. """
        return get_gsod_daily_temp_data_cache_key(self.usaf_id, year)

    def get_tmy3_hourly_temp_data_cache_key(self):
        """ Get key used to cache TMY3 weather-normalized temperature data. """
        return get_tmy3_hourly_temp_data_cache_key(self.usaf_id)

    def get_cz2010_hourly_temp_data_cache_key(self):
        """ Get key used to cache CZ2010 weather-normalized temperature data. """
        return get_cz2010_hourly_temp_data_cache_key(self.usaf_id)

    # is cached data expired? boolean. true if expired or not in cache
    def cached_isd_hourly_temp_data_is_expired(self, year):
        """ Return True if cache of resampled hourly ISD temperature data has expired or does not exist for the given year. """
        return cached_isd_hourly_temp_data_is_expired(self.usaf_id, year)

    def cached_isd_daily_temp_data_is_expired(self, year):
        """ Return True if cache of resampled daily ISD temperature data has expired or does not exist for the given year. """
        return cached_isd_daily_temp_data_is_expired(self.usaf_id, year)

    def cached_gsod_daily_temp_data_is_expired(self, year):
        """ Return True if cache of resampled daily GSOD temperature data has expired or does not exist for the given year. """
        return cached_gsod_daily_temp_data_is_expired(self.usaf_id, year)

    # check if data is available and delete data in the cache if it's expired
    def validate_isd_hourly_temp_data_cache(self, year):
        """ Delete cached resampled hourly ISD temperature data if it has expired for the given year. """
        return validate_isd_hourly_temp_data_cache(self.usaf_id, year)

    def validate_isd_daily_temp_data_cache(self, year):
        """ Delete cached resampled daily ISD temperature data if it has expired for the given year. """
        return validate_isd_daily_temp_data_cache(self.usaf_id, year)

    def validate_gsod_daily_temp_data_cache(self, year):
        """ Delete cached resampled daily GSOD temperature data if it has expired for the given year. """
        return validate_gsod_daily_temp_data_cache(self.usaf_id, year)

    def validate_tmy3_hourly_temp_data_cache(self):
        """ Check if TMY3 data exists in cache. """
        return validate_tmy3_hourly_temp_data_cache(self.usaf_id)

    def validate_cz2010_hourly_temp_data_cache(self):
        """ Check if CZ2010 data exists in cache. """
        return validate_cz2010_hourly_temp_data_cache(self.usaf_id)

    # pandas time series to json
    def serialize_isd_hourly_temp_data(self, ts):
        """ Serialize resampled hourly ISD pandas time series as JSON for caching. """
        return serialize_isd_hourly_temp_data(ts)

    def serialize_isd_daily_temp_data(self, ts):
        """ Serialize resampled daily ISD pandas time series as JSON for caching. """
        return serialize_isd_daily_temp_data(ts)

    def serialize_gsod_daily_temp_data(self, ts):
        """ Serialize resampled daily GSOD pandas time series as JSON for caching. """
        return serialize_gsod_daily_temp_data(ts)

    def serialize_tmy3_hourly_temp_data(self, ts):
        """ Serialize hourly TMY3 pandas time series as JSON for caching. """
        return serialize_tmy3_hourly_temp_data(ts)

    def serialize_cz2010_hourly_temp_data(self, ts):
        """ Serialize hourly CZ2010 pandas time series as JSON for caching. """
        return serialize_cz2010_hourly_temp_data(ts)

    # json to pandas time series
    def deserialize_isd_hourly_temp_data(self, data):
        """ Deserialize JSON representation of resampled hourly ISD into pandas time series. """
        return deserialize_isd_hourly_temp_data(data)

    def deserialize_isd_daily_temp_data(self, data):
        """ Deserialize JSON representation of resampled daily ISD into pandas time series. """
        return deserialize_isd_daily_temp_data(data)

    def deserialize_gsod_daily_temp_data(self, data):
        """ Deserialize JSON representation of resampled daily GSOD into pandas time series. """
        return deserialize_gsod_daily_temp_data(data)

    def deserialize_tmy3_hourly_temp_data(self, data):
        """ Deserialize JSON representation of hourly TMY3 into pandas time series. """
        return deserialize_isd_hourly_temp_data(data)

    def deserialize_cz2010_hourly_temp_data(self, data):
        """ Deserialize JSON representation of hourly CZ2010 into pandas time series. """
        return deserialize_cz2010_hourly_temp_data(data)

    # return pandas time series of data from cache
    def read_isd_hourly_temp_data_from_cache(self, year):
        """ Get cached version of resampled hourly ISD temperature data for given year. """
        return read_isd_hourly_temp_data_from_cache(self.usaf_id, year)

    def read_isd_daily_temp_data_from_cache(self, year):
        """ Get cached version of resampled daily ISD temperature data for given year. """
        return read_isd_daily_temp_data_from_cache(self.usaf_id, year)

    def read_gsod_daily_temp_data_from_cache(self, year):
        """ Get cached version of resampled daily GSOD temperature data for given year. """
        return read_gsod_daily_temp_data_from_cache(self.usaf_id, year)

    def read_tmy3_hourly_temp_data_from_cache(self):
        """ Get cached version of hourly TMY3 temperature data. """
        return read_tmy3_hourly_temp_data_from_cache(self.usaf_id)

    def read_cz2010_hourly_temp_data_from_cache(self):
        """ Get cached version of hourly TMY3 temperature data. """
        return read_cz2010_hourly_temp_data_from_cache(self.usaf_id)

    # write pandas time series of data to cache for a particular year
    def write_isd_hourly_temp_data_to_cache(self, year, ts):
        """ Write resampled hourly ISD temperature data to cache for given year. """
        return write_isd_hourly_temp_data_to_cache(self.usaf_id, year, ts)

    def write_isd_daily_temp_data_to_cache(self, year, ts):
        """ Write resampled daily ISD temperature data to cache for given year. """
        return write_isd_daily_temp_data_to_cache(self.usaf_id, year, ts)

    def write_gsod_daily_temp_data_to_cache(self, year, ts):
        """ Write resampled daily GSOD temperature data to cache for given year. """
        return write_gsod_daily_temp_data_to_cache(self.usaf_id, year, ts)

    def write_tmy3_hourly_temp_data_to_cache(self, ts):
        """ Write hourly TMY3 temperature data to cache for given year. """
        return write_tmy3_hourly_temp_data_to_cache(self.usaf_id, ts)

    def write_cz2010_hourly_temp_data_to_cache(self, ts):
        """ Write hourly CZ2010 temperature data to cache for given year. """
        return write_cz2010_hourly_temp_data_to_cache(self.usaf_id, ts)

    # delete cached data for a particular year
    def destroy_cached_isd_hourly_temp_data(self, year):
        """ Remove cached resampled hourly ISD temperature data to cache for given year. """
        return destroy_cached_isd_hourly_temp_data(self.usaf_id, year)

    def destroy_cached_isd_daily_temp_data(self, year):
        """ Remove cached resampled daily ISD temperature data to cache for given year. """
        return destroy_cached_isd_daily_temp_data(self.usaf_id, year)

    def destroy_cached_gsod_daily_temp_data(self, year):
        """ Remove cached resampled daily GSOD temperature data to cache for given year. """
        return destroy_cached_gsod_daily_temp_data(self.usaf_id, year)

    def destroy_cached_tmy3_hourly_temp_data(self):
        """ Remove cached hourly TMY3 temperature data to cache. """
        return destroy_cached_tmy3_hourly_temp_data(self.usaf_id)

    def destroy_cached_cz2010_hourly_temp_data(self):
        """ Remove cached hourly CZ2010 temperature data to cache. """
        return destroy_cached_cz2010_hourly_temp_data(self.usaf_id)

    # load data either from cache if valid or directly from source
    def load_isd_hourly_temp_data_cached_proxy(self, year, fetch_from_web=True):
        """ Load resampled hourly ISD temperature data from cache, or if it is expired or hadn't been cached, fetch from FTP for given year. """
        return load_isd_hourly_temp_data_cached_proxy(
            self.usaf_id, year, fetch_from_web
        )

    def load_isd_daily_temp_data_cached_proxy(self, year, fetch_from_web=True):
        """ Load resampled daily ISD temperature data from cache, or if it is expired or hadn't been cached, fetch from FTP for given year. """
        return load_isd_daily_temp_data_cached_proxy(self.usaf_id, year, fetch_from_web)

    def load_gsod_daily_temp_data_cached_proxy(self, year, fetch_from_web=True):
        """ Load resampled daily GSOD temperature data from cache, or if it is expired or hadn't been cached, fetch from FTP for given year. """
        return load_gsod_daily_temp_data_cached_proxy(
            self.usaf_id, year, fetch_from_web
        )

    def load_tmy3_hourly_temp_data_cached_proxy(self, fetch_from_web=True):
        """ Load hourly TMY3 temperature data from cache, or if it is expired or hadn't been cached, fetch from NREL. """
        return load_tmy3_hourly_temp_data_cached_proxy(self.usaf_id, fetch_from_web)

    def load_cz2010_hourly_temp_data_cached_proxy(self, fetch_from_web=True):
        """ Load hourly CZ2010 temperature data from cache, or if it is expired or hadn't been cached, fetch from URL. """
        return load_cz2010_hourly_temp_data_cached_proxy(self.usaf_id, fetch_from_web)

    # main interface: load data from start date to end date
    def load_isd_hourly_temp_data(
        self,
        start,
        end,
        read_from_cache=True,
        write_to_cache=True,
        fetch_from_web=True,
        error_on_missing_years=True,
    ):
        """ Load resampled hourly ISD temperature data from start date to end date (inclusive).

        This is the primary convenience method for loading resampled hourly ISD temperature data.

        Parameters
        ----------
        start : datetime.datetime
            The earliest date from which to load data.
        end : datetime.datetime
            The latest date until which to load data.
        read_from_cache : bool
            Whether or not to load data from cache.
        fetch_from_web : bool
            Whether or not to fetch data from ftp.
        write_to_cache : bool
            Whether or not to write newly loaded data to cache.
        """
        return load_isd_hourly_temp_data(
            self.usaf_id,
            start,
            end,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
            error_on_missing_years=error_on_missing_years,
        )

    def load_isd_daily_temp_data(
        self, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
    ):
        """ Load resampled daily ISD temperature data from start date to end date (inclusive).

        This is the primary convenience method for loading resampled daily ISD temperature data.

        Parameters
        ----------
        start : datetime.datetime
            The earliest date from which to load data.
        end : datetime.datetime
            The latest date until which to load data.
        read_from_cache : bool
            Whether or not to load data from cache.
        fetch_from_web : bool
            Whether or not to fetch data from ftp.
        write_to_cache : bool
            Whether or not to write newly loaded data to cache.
        """
        return load_isd_daily_temp_data(
            self.usaf_id,
            start,
            end,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
        )

    def load_gsod_daily_temp_data(
        self, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
    ):
        """ Load resampled daily GSOD temperature data from start date to end date (inclusive).

        This is the primary convenience method for loading resampled daily GSOD temperature data.

        Parameters
        ----------
        start : datetime.datetime
            The earliest date from which to load data.
        end : datetime.datetime
            The latest date until which to load data.
        read_from_cache : bool
            Whether or not to load data from cache.
        write_to_cache : bool
            Whether or not to write newly loaded data to cache.
        fetch_from_web : bool
            Whether or not to fetch data from ftp.
        """
        return load_gsod_daily_temp_data(
            self.usaf_id,
            start,
            end,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
        )

    def load_tmy3_hourly_temp_data(
        self, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
    ):
        """ Load hourly TMY3 temperature data from start date to end date (inclusive).

        This is the primary convenience method for loading hourly TMY3 temperature data.

        Parameters
        ----------
        start : datetime.datetime
            The earliest date from which to load data.
        end : datetime.datetime
            The latest date until which to load data.
        read_from_cache : bool
            Whether or not to load data from cache.
        write_to_cache : bool
            Whether or not to write newly loaded data to cache.
        fetch_from_web : bool
            Whether or not to fetch data from ftp.
        """
        return load_tmy3_hourly_temp_data(
            self.usaf_id,
            start,
            end,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
        )

    def load_cz2010_hourly_temp_data(
        self, start, end, read_from_cache=True, write_to_cache=True, fetch_from_web=True
    ):
        """ Load hourly CZ2010 temperature data from start date to end date (inclusive).

        This is the primary convenience method for loading hourly CZ2010 temperature data.

        Parameters
        ----------
        start : datetime.datetime
            The earliest date from which to load data.
        end : datetime.datetime
            The latest date until which to load data.
        read_from_cache : bool
            Whether or not to load data from cache.
        write_to_cache : bool
            Whether or not to write newly loaded data to cache.
        fetch_from_web : bool
            Whether or not to fetch data from ftp.
        """
        return load_cz2010_hourly_temp_data(
            self.usaf_id,
            start,
            end,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
            fetch_from_web=fetch_from_web,
        )

    # load all cached data for this station
    def load_cached_isd_hourly_temp_data(self):
        """ Load all cached resampled hourly ISD temperature data. """
        return load_cached_isd_hourly_temp_data(self.usaf_id)

    def load_cached_isd_daily_temp_data(self):
        """ Load all cached resampled daily ISD temperature data. """
        return load_cached_isd_daily_temp_data(self.usaf_id)

    def load_cached_gsod_daily_temp_data(self):
        """ Load all cached resampled daily GSOD temperature data. """
        return load_cached_gsod_daily_temp_data(self.usaf_id)

    def load_cached_tmy3_hourly_temp_data(self):
        """ Load all cached hourly TMY3 temperature data (the year is set to 1900) """
        return load_cached_tmy3_hourly_temp_data(self.usaf_id)

    def load_cached_cz2010_hourly_temp_data(self):
        """ Load all cached hourly TMY3 temperature data (the year is set to 1900) """
        return load_cached_cz2010_hourly_temp_data(self.usaf_id)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_isd_station_metadata_table_count count"] = 3874

snapshots["test_isd_file_metadata_table_count count"] = 29668

snapshots["test_isd_station_metadata_table_content data"] = {
    "ba_climate_zone": "Very Cold",
    "ca_climate_zone": None,
    "elevation": "+0031.4",
    "icao_code": "PATO",
    "iecc_climate_zone": "7",
    "iecc_moisture_regime": None,
    "latitude": "+60.785",
    "longitude": "-148.839",
    "name": "PORTAGE GLACIER",
    "quality": "high",
    "recent_wban_id": "26492",
    "state": "AK",
    "usaf_id": "700001",
    "wban_ids": "26492,99999",
}

snapshots["test_isd_file_metadata_table_content data"] = {
    "usaf_id": "690090",
    "wban_id": "99999",
    "year": "2008",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import ftplib
from io import BytesIO
import logging
import os
import requests
import sqlite3

from .cache import KeyValueStore

logger = logging.getLogger(__name__)

__all__ = ("noaa_ftp_connection_proxy", "metadata_db_connection_proxy")


def _get_noaa_ftp_connection(n_tries=5, timeout=60):  # pragma: no cover
    host = "ftp.ncdc.noaa.gov"
    for i in range(n_tries):
        # attempt anonymous connection
        try:
            ftp = ftplib.FTP(host, timeout=timeout)
            ftp.login()  # default u='anonymous' p='anonymous@'
            logger.info("Connected to {}.".format(host))
            return ftp
        except ftplib.all_errors as e:
            logger.warn(
                "Failed attempt ({} of {}) to connect to {}:\n{}".format(
                    i + 1, n_tries, host, e
                )
            )

    raise RuntimeError("Could not connect to {}.".format(host))


class NOAAFTPConnectionProxy(object):
    def __init__(self):
        self._connection = None

    def get_connection(self):  # pragma: no cover
        if self._connection is None:
            self._connection = _get_noaa_ftp_connection()
        return self._connection

    def reconnect(self):  # pragma: no cover
        if self._connection is not None:
            self._connection.close()
            self._connection = None
        return self.get_connection()

    def read_file_as_bytes(self, filename):  # pragma: no cover
        ftp = self.get_connection()

        bytes_string = BytesIO()
        try:
            try:
                ftp.retrbinary("RETR {}".format(filename), bytes_string.write)
            except (ftplib.error_temp, ftplib.error_perm, EOFError, IOError) as e:
                # Bad connection. attempt to reconnect.
                logger.warn(
                    "Failed RETR {}:\n{}\n" "Attempting reconnect.".format(filename, e)
                )
                ftp = self.reconnect()
                ftp.retrbinary("RETR {}".format(filename), bytes_string.write)
        except Exception as e:
            logger.warn(
                "Failed RETR {}:\n{}\n" "Not attempting reconnect.".format(filename, e)
            )
            return None

        bytes_string.seek(0)
        logger.info("Successfully retrieved ftp://ftp.ncdc.noaa.gov{}".format(filename))
        return bytes_string


class MetadataDBConnectionProxy(object):
    def __init__(self):
        root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        path = os.path.join(root_dir, "eeweather", "resources")
        self.db_path = os.path.join(path, "metadata.db")

    def get_connection(self):
        return sqlite3.connect(self.db_path)

    def reset_database(self):  # pragma: no cover
        return self.get_connection()


class KeyValueStoreProxy(object):
    def __init__(self):
        self._store = None

    def get_store(self):  # pragma: no cover
        if self._store is None:
            self._store = KeyValueStore()
        return self._store


# Use proxies for lazy loading, abstraction
noaa_ftp_connection_proxy = NOAAFTPConnectionProxy()
metadata_db_connection_proxy = MetadataDBConnectionProxy()
key_value_store_proxy = KeyValueStoreProxy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json
import subprocess

import click

from . import (
    get_isd_station_metadata as _get_isd_station_metadata,
    get_isd_file_metadata as _get_isd_file_metadata,
    get_isd_filenames as _get_isd_filenames,
    get_gsod_filenames as _get_gsod_filenames,
)
from .exceptions import UnrecognizedUSAFIDError

from .database import build_metadata_db, inspect_metadata_db


@click.group()
def cli():
    """Example usage

    See station metadata:

    \b
        $ eeweather inspect_isd_station 722880
        {
          "usaf_id": "722880",
          "wban_ids": "23152,99999",
          "recent_wban_id": "23152",
          "name": "BURBANK-GLENDALE-PASA ARPT",
          "latitude": "+34.201",
          "longitude": "-118.358",
          "elevation": "+0236.2"
        }

    See station file data:

    \b
        $ eeweather inspect_isd_file_years 722880
        {...}

    See station file names for ISD:

    \b
        $ eeweather inspect_isd_filenames 722880 2017
        ftp://ftp.ncdc.noaa.gov/pub/data/noaa/2017/722880-23152-2017.gz

    Or for GSOD:

    \b
        $ eeweather inspect_gsod_filenames 722880 2017
        ftp://ftp.ncdc.noaa.gov/pub/data/gsod/2017/722880-23152-2017.op.gz

    Rebuild metadata db from primary source files:

    \b
        $ eeweather rebuild_db

    Inspect metadata db using sqlite3 CLI:

    \b
        $ eeweather inspect_db


    """
    pass  # pragma: no cover


@cli.command()
@click.argument("usaf_id")
def inspect_isd_station(usaf_id):
    metadata = _get_isd_station_metadata(usaf_id)
    click.echo(json.dumps(metadata, indent=2))


@cli.command()
@click.argument("usaf_id")
def inspect_isd_file_years(usaf_id):
    metadata = _get_isd_file_metadata(usaf_id)
    click.echo(json.dumps(metadata, indent=2))


@cli.command()
@click.argument("usaf_id")
@click.argument("year")
def inspect_isd_filenames(usaf_id, year):
    filenames = _get_isd_filenames(usaf_id, year, with_host=True)
    for f in filenames:
        click.echo(f)


@cli.command()
@click.argument("usaf_id")
@click.argument("year")
def inspect_gsod_filenames(usaf_id, year):
    filenames = _get_gsod_filenames(usaf_id, year, with_host=True)
    for f in filenames:
        click.echo(f)


@cli.command()
@click.option("--zcta-geometry/--no-zcta-geometry", default=False)
@click.option(
    "--iecc-climate-zone-geometry/--no-iecc-climate-zone-geometry", default=True
)
@click.option(
    "--iecc-moisture-regime-geometry/--no-iecc-moisture-regime-geometry", default=True
)
@click.option("--ba-climate-zone-geometry/--no-ba-climate-zone-geometry", default=True)
@click.option("--ca-climate-zone-geometry/--no-ca-climate-zone-geometry", default=True)
def rebuild_db(
    zcta_geometry,
    iecc_climate_zone_geometry,
    iecc_moisture_regime_geometry,
    ba_climate_zone_geometry,
    ca_climate_zone_geometry,
):
    build_metadata_db(  # pragma: no cover
        zcta_geometry,
        iecc_climate_zone_geometry,
        iecc_moisture_regime_geometry,
        ba_climate_zone_geometry,
        ca_climate_zone_geometry,
    )


@cli.command()
def inspect_db():
    inspect_metadata_db()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from .connections import metadata_db_connection_proxy

__all__ = ("get_zcta_ids", "get_isd_station_usaf_ids")


def get_zcta_ids(state=None):
    """ Get ids of all supported ZCTAs, optionally by state.

    Parameters
    ----------
    state : str, optional
        Select zipcodes only from this state or territory, given as 2-letter
        abbreviation (e.g., ``'CA'``, ``'PR'``).

    Returns
    -------
    results : list of str
        List of all supported selected ZCTA IDs.
    """
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()

    if state is None:
        cur.execute(
            """
          select zcta_id from zcta_metadata
        """
        )
    else:
        cur.execute(
            """
          select zcta_id from zcta_metadata where state = ?
        """,
            (state,),
        )
    return [row[0] for row in cur.fetchall()]


def get_isd_station_usaf_ids(state=None):
    """ Get USAF IDs of all supported ISD stations, optionally by state.

    Parameters
    ----------
    state : str, optional
        Select ISD station USAF IDs only from this state or territory, given
        as 2-letter abbreviation (e.g., ``'CA'``, ``'PR'``).

    Returns
    -------
    results : list of str
        List of all supported selected ISD station USAF IDs.
    """
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()

    if state is None:
        cur.execute(
            """
          select usaf_id from isd_station_metadata
        """
        )
    else:
        cur.execute(
            """
          select usaf_id from isd_station_metadata where state = ?
        """,
            (state,),
        )
    return [row[0] for row in cur.fetchall()]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""


class lazy_property(object):
    """
    Meant to be used for lazy evaluation of an object attribute.
    Property should represent non-mutable data, as it replaces itself.

    e.g.,

    class Test(object):

        @lazy_property
        def results(self):
            calcs = 1  # Do a lot of calculation here
            return calcs

    from https://stackoverflow.com/a/6849299/1965736
    """

    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        value = self.fget(obj)
        setattr(obj, self.func_name, value)
        return value
